'''
The MIT License (MIT)

Copyright (c) 2014 Tony Walker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import uuid
import nanomsg
import logging

from .encoder import MsgPackEncoder
from .error import ClientError


class Client(object):
    """ A requester client """

    def __init__(self, addr, encoder=None, socket=None):
        self.addr = addr
        self.encoder = encoder or MsgPackEncoder()
        self.sock = socket if socket else nanomsg.Socket(nanomsg.REQ)
        self.sock.connect(self.addr)

    def build_payload(self, method, args):
        ref = str(uuid.uuid4())
        return (method, args, ref)

    def encode(self, payload):
        return self.encoder.encode(payload)

    def decode(self, msg):
        return self.encoder.decode(msg)

    def call(self, method, *args):
        payload = self.build_payload(method, args)
        self.sock.send(self.encode(payload))
        res = self.decode(self.sock.recv())
        assert payload[2] == res['ref']
        return res['result'], res['error']


class PubClient(Client):
    """ A publisher client """

    def __init__(self, addr, encoder=None):
        socket = nanomsg.Socket(nanomsg.PUB)
        super(PubClient, self).__init__(addr, encoder, socket)

    def call(self, method, *args):
        raise ClientError('Operation not allowed on this type of client')

    def build_payload(self, tag, data):
        payload = bytes(tag.encode('utf-8')) + self.encode(data)
        return payload

    def publish(self, tag, data):
        payload = self.build_payload(tag, data)
        self.sock.send(payload)



'''
The MIT License (MIT)

Copyright (c) 2014 Tony Walker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import sys
import signal
import nanomsg
import logging

from .encoder import MsgPackEncoder
from .error import ServiceError


class Service(object):

    def __init__(self, addr, encoder=None, socket=None):
        self.addr = addr
        self.encoder = encoder or MsgPackEncoder()
        self.methods = {}
        self.descriptions = {}
        self.sock = socket if socket else nanomsg.Socket(nanomsg.REP)
        self.sock.bind(self.addr)

    def recv(self):
        payload = self.sock.recv()
        decoded = self.encoder.decode(payload)
        return decoded

    def send(self, response):
        response = self.encoder.encode(response)
        self.sock.send(response)

    def execute(self, method, args, ref):
        response = {'result': None, 'error': None, 'ref': ref}

        # Execute the requested method
        fun = self.methods.get(method)
        if not fun:
            response['error'] = 'Method `{}` not found'.format(method)
        else:
            try:
                response['result'] = fun(*args)
            except Exception as e:
                logging.error(e, exc_info=1)
                response['error'] = str(e)
        return response

    def register(self, name, fun, description=None):
        self.methods[name] = fun
        self.descriptions[name] = description

    def process(self):
        payload = self.recv()
        logging.debug(
            '* Server received payload: {}'.format(payload))
        try:
            method, args, ref = payload
            response = self.execute(method, args, ref)
        except Exception as e:
            logging.error(e, exc_info=1)
        else:
            self.send(response)

    def start(self):
        """ Start and listen for calls """

        signal.signal(signal.SIGINT, self.stop)
        logging.info('* Service started on {}'.format(self.addr))

        while True:
            self.process()

    def stop(self, signal=None, frame=None):
        logging.info('* Shutting down service ...')
        self.sock.close()
        sys.exit(0)


class SubService(Service):
    """ A service which subscribes """

    def __init__(self, addr, encoder=None):
        socket = nanomsg.Socket(nanomsg.SUB)
        super(SubService, self).__init__(addr, encoder, socket)

    def get_fun_and_data(self, msg):
        for name in self.methods:
            tag = bytes(name.encode('utf-8'))
            if msg.startswith(tag):
                fun = self.methods.get(name)
                data = self.encoder.decode(msg[len(tag):])
                return fun, data
        return None, None

    def register(self, name, fun):
        raise ServiceError('Operation not allowed on this type of service')

    def subscribe(self, name, fun):
        super(SubService, self).register(name, fun)
        self.sock.set_string_option(nanomsg.SUB, nanomsg.SUB_SUBSCRIBE, name)

    def process(self):
        msg = self.sock.recv()
        fun, data = self.get_fun_and_data(msg)

        result = None
        try:
            result = fun(data)
        except Exception as e:
            logging.error(e, exc_info=1)

        # Return result to check successful execution
        # of `fun` when testing
        return result



#!/usr/bin/env python

import os
import platform
import sys
import warnings

try:
    import setuptools
    from setuptools import setup
except ImportError:
    setuptools = None
    from distutils.core import setup


readme_file = 'README.md'
try:
    import pypandoc
    long_description = pypandoc.convert(readme_file, 'rst')
except (ImportError, OSError) as e:
    print('No pypandoc or pandoc: %s' % (e,))
    with open(readme_file) as fh:
        long_description = fh.read()

with open('./nanoservice/version.py') as fh:
    for line in fh:
        if line.startswith('VERSION'):
            version = line.split('=')[1].strip().strip("'")

setup(
    name='nanoservice',
    version=version,
    packages = ['nanoservice'],
    author='Tony Walker',
    author_email='walkr.walkr@gmail.com',
    url='https://github.com/walkr/nanoservice',
    license='MIT',
    description='nanoservice is a small Python library for writing lightweight networked services using nanomsg',
    long_description=long_description,
    install_requires=[
        'msgpack-python',
        'nanomsg',
        'nose',
    ],
    dependency_links=[
        'git+https://github.com/tonysimpson/nanomsg-python.git@master#egg=nanomsg',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        ],
)
